#include "igk_motor.h"


// Protocol version
#define PROTOCOL_VERSION1               1.0                 // See which protocol version is used in the Dynamixel
#define PROTOCOL_VERSION2               2.0

// Default setting
#if defined(__OPENCR__) 
#define DEVICENAME                      "/dev/OpenCR"       // Device name not used on OpenCR
#elif defined(__OPENCM904__)
#define DEVICENAME                      "3"                 // Default to external (OpenCM485 expansion) on OpenCM9.04
#endif

#define CMD_SERIAL                      Serial1              // 串口1用于打印驱动信息

char *dev_name = (char*)DEVICENAME;
/***********************************************/
//PID变量初始化
/***********************************************/
Pid PID_M1;//电机PID
signed char error = 0;//此导航偏差
int data_cdh = 0;//此导航数据
int speedx = 100;//基速度
dynamixel::PacketHandler *packetHandler2;
dynamixel::PortHandler   *portHandler;
can_message_t rx_msg;
int tb3_baud = -1;
void init_bsp()
{
    CMD_SERIAL.begin(115200);
    //初始化电机端口
    packetHandler2 = dynamixel::PacketHandler::getPacketHandler(PROTOCOL_VERSION2);
    portHandler    = dynamixel::PortHandler::getPortHandler(dev_name);
    portHandler->openPort();
    if (portHandler->openPort())
    {
        tb3_baud = portHandler->getBaudRate();
    }
    else
    {
        CMD_SERIAL.printf("Failed to open the port!\n");
        CMD_SERIAL.printf("Press any key to terminate...\n");
        while (1);
    }
    //can
    //初始化CAN总线
    if (CanBus.begin(_DEF_CAN_BAUD_500K, CAN_STD_FORMAT) == false)
    {
       Serial.println("CAN open fail!!");
    }
    else
    {
        CanBus.configFilter(0x123, 0);
        CanBus.attachRxInterrupt(canRxHandlerTemplate);
    }
    //设置电机通信波特率
    portHandler->setBaudRate(1000000);
    //初始化电机PID
    PID_Init();
}
int control(char cmd,int speed_defualt,int speed_max)
{
    signed char inc = 0;
    int speed_m1 = 0,speed_m2 = 0;
    if(cmd == 'G')
    {
        //出轨停止
        if(data_cdh == 0)
        {
            Motor_stop(1);
            Motor_stop(2); 
            Motor_stop(3);
            Motor_stop(4);   
        }
        //到达终点停止
        else if(data_cdh == 0xffff)
        {
          Motor_stop(1);
          Motor_stop(2); 
          Motor_stop(3);
          Motor_stop(4);  
          return 1;
        }
        else
        {
            PosPIDCalc(&PID_M1,error);
    //       CMD_SERIAL.printf("error=%d\n",error);
    //        CMD_SERIAL.printf("PID_M1.result=%d\n",PID_M1.result);
            speed_m1 = speed_defualt+PID_M1.result;
            speed_m2 = speed_defualt-PID_M1.result;
            if(speed_m1>speed_max)
            {
                speed_m1 = speed_max;
            }
            else if(speed_m1<-speed_max)
            {
                speed_m1 = -speed_max;
            }
            if(speed_m2>speed_max)
            {
                speed_m2 = speed_max;
            }
            else if(speed_m2<-speed_max)
            {
                speed_m2 = -speed_max;
            }
            Motor(1,speed_m1);
            Motor(2,speed_m2);  
            Motor(3,speed_m1);
            Motor(4,speed_m2);    
        }      
    }
    else if(cmd == 'S')
    {
      Motor_stop(1);
      Motor_stop(2); 
      Motor_stop(3);
      Motor_stop(4);   
    }
    return 0;         
}
void Motor_stop(uint8_t id)
{
    write(portHandler, packetHandler2, id, 104, 4, 0); 
    write(portHandler, packetHandler2, id, 64, 1, 0);  
}
void Motor(uint8_t id,int speed)
{
  uint32_t pre_time;
  uint8_t  toggle = 0; 
  // We run at 1000000

  uint16_t model_number;
  //如果速度为0关闭使能
  if(speed == 0)
    write(portHandler, packetHandler2, id, 64, 1, 0);
  else
  {
    write(portHandler, packetHandler2, id, 64, 1, 1);
    write(portHandler, packetHandler2, id, 104, 4, speed);
  }
}
//得到磁导航偏差
void canRxHandlerTemplate(can_message_t *arg)
{
  uint32_t id, i;
  if(CanBus.readMessage(&rx_msg))
  {
    error = (signed char)(rx_msg.data[3]);
    data_cdh = (rx_msg.data[1]<<8)|rx_msg.data[2];
  }
}
void PID_Init(void)
{
    PID_M1.Kp = 5;
    PID_M1.Ki = 0;
    PID_M1.Kd = 0;
    PID_M1.LastError = 0;
    PID_M1.Error = 0;     //误差
    PID_M1.PrevError = 0;
    PID_M1.SumError = 0;  //累计误差
    PID_M1.SetTarget = 0; //设定目标
    PID_M1.max = 100;
    PID_M1.min = 0;
}
/******************************************************/
//位置式PID计算
//公式:U(k) = Kp*[E(n)] + Ki*SumError + Kd*[E(n)-E(n-1)]
//参数说明:Error = E(n) 
/*******************************************************/
void PosPIDCalc(Pid *vpid,int NextTarget)
{
  float Error;

    vpid->Error = vpid->SetTarget - NextTarget;     //E(n)当前偏差E(n)  
    
    if(vpid->result > vpid->max)
    {
        if(vpid->Error<0)
        {
            vpid->SumError +=vpid->Error;                   //SumError误差和             
        }        
    }
    else if(vpid->result <= vpid->min)
    {
        if(vpid->Error>0)
        {
            vpid->SumError +=vpid->Error;                   //SumError误差和             
        }         
    }
    else
    {
            vpid->SumError +=vpid->Error;                   //SumError误差和          
    }
    Error = vpid->Error - vpid->LastError;          //[E(n)-E(n-1)]微分

    vpid->LastError = vpid->Error;                                  //更新偏差值

    vpid->result = vpid->Kp*vpid->Error + vpid->Ki*vpid->SumError + vpid->Kd*Error;            
    
}

void write(dynamixel::PortHandler *portHandler, dynamixel::PacketHandler *packetHandler, uint8_t id, uint16_t addr, uint16_t length, uint32_t value)
{
  uint8_t dxl_error = 0;
  int dxl_comm_result = COMM_TX_FAIL;

  if (length == 1)
  {
    dxl_comm_result = packetHandler->write1ByteTxRx(portHandler, id, addr, (uint8_t)value, &dxl_error);
  }
  else if (length == 2)
  {
    dxl_comm_result = packetHandler->write2ByteTxRx(portHandler, id, addr, (uint16_t)value, &dxl_error);
  }
  else if (length == 4)
  {
    dxl_comm_result = packetHandler->write4ByteTxRx(portHandler, id, addr, (uint32_t)value, &dxl_error);
  }

  if (dxl_comm_result == COMM_SUCCESS)
  {
    if (dxl_error != 0) CMD_SERIAL.println(packetHandler->getRxPacketError(dxl_error));
  }
  else
  {
    CMD_SERIAL.println(packetHandler->getTxRxResult(dxl_error));
    CMD_SERIAL.println("Fail to write!");
  }
}
